import os
from typing import Any, Iterable, TypedDict

from . import to_array


class SeratoBinDb:
    FIELDNAMES = {
        # Database & Crate
        "vrsn": "Version",
        "otrk": "Track",
        # Database
        "ttyp": "File Type",
        "pfil": "File Path",
        "tsng": "Song Title",
        "tart": "Artist",
        "talb": "Album",
        "tgen": "Genre",
        "tlen": "Length",
        "tbit": "Bitrate",
        "tsmp": "Sample Rate",
        "tsiz": "Size",
        "tbpm": "BPM",
        "tkey": "Key",
        "tart": "Artist",
        "utme": "File Time",
        "tgrp": "Grouping",
        "tlbl": "Publisher",
        "tcmp": "Composer",
        "ttyr": "Year",
        # Serato stuff
        "tadd": "Date added",
        "uadd": "Date added",
        "bbgl": "Beatgrid Locked",
        "bcrt": "Corrupt",
        "bmis": "Missing",
        # Crates
        "osrt": "Sorting",
        "brev": "Reverse Order",
        "ovct": "Column",
        "tvcn": "Column Name",
        "tvcw": "Column Width",
        "ptrk": "Track Path",
    }
    FIELDNAME_KEYS = list(FIELDNAMES.keys())
    TRACK_FIELD = "otrk"

    class DataTypeError(Exception):
        def __init__(
            self, value: Any, expected_type: type | Iterable[type], field: str | None
        ):
            super().__init__(
                f"value must be {' or '.join(e.__name__ for e in to_array(expected_type))} when field is {field} (type: {type(value).__name__}) (value: {str(value)})"
            )

    @staticmethod
    def get_field_name(field: str):
        return SeratoBinDb.FIELDNAMES.get(field, "Unknown Field")

    @staticmethod
    def _get_type(field: str) -> str:
        # vrsn field has no type_id, but contains text ("t")
        return "t" if field == "vrsn" else field[0]

    @staticmethod
    def _check_valid_field(field: str):
        if field not in SeratoBinDb.FIELDNAME_KEYS:
            raise ValueError(
                f"invalid field: {field} must be one of: {str(SeratoBinDb.FIELDNAME_KEYS)}\n(see {__file__} for what these keys map to)"
            )

    @staticmethod
    def remove_drive_from_filepath(filepath: str) -> tuple[str, str]:
        drive, filepath = os.path.splitdrive(filepath)
        return os.path.normpath(filepath).lstrip("\\"), drive

    class FieldObj(TypedDict):
        field: str

    @staticmethod
    def _check_rule_fields(rules: Iterable[FieldObj]):
        all_field_names = [rule["field"] for rule in rules]
        uniq_field_names = list(set(all_field_names))
        assert len(list(rules)) == len(
            uniq_field_names
        ), f"must only have 1 function per field. fields passed: {str(sorted(all_field_names))}"
        for field in uniq_field_names:
            SeratoBinDb._check_valid_field(field)



import io
import struct

import mutagen.id3
from mutagen.id3._frames import GEOB

_MutagenTagFile = mutagen.id3.ID3FileType | mutagen.id3.ID3


def get_geob(tagfile: _MutagenTagFile, geob_key: str) -> bytes:
    geob_key = f"GEOB:{geob_key}"
    try:
        return tagfile[geob_key].data
    except KeyError:
        raise KeyError(f'File is missing "{geob_key}" tag')


def tag_geob(tagfile: _MutagenTagFile, geob_key: str, data: bytes):
    tagfile[f"GEOB:{geob_key}"] = GEOB(
        encoding=0,
        mime="application/octet-stream",
        desc=geob_key,
        data=data,
    )


def del_tag(tagfile: _MutagenTagFile, key: str):
    """
    Returns True if was deleted, False if no change
    """
    if key in tagfile:
        del tagfile[key]
        return True
    return False


def del_geob(tagfile: _MutagenTagFile, geob_key: str):
    """
    Returns True if was deleted, False if no change
    """
    return del_tag(tagfile, f"GEOB:{geob_key}")


VERSION_FORMAT = "BB"
VersionType = tuple[int, int]


def check_version(given: bytes, expected: VersionType):
    given_version = struct.unpack(VERSION_FORMAT, given)
    if given_version != expected:
        raise ValueError(
            f"Wrong version. Expected: {str(expected)} Given: {str(given_version)}"
        )


def pack_version(version: VersionType):
    return struct.pack(VERSION_FORMAT, *version)



#!/usr/bin/python
# This is from this repo: https://github.com/sharst/seratopy
import logging
import os
import struct
import sys
from typing import Generator, TypedDict, Union, overload

if __package__ is None:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from serato_tools.utils.database import SeratoBinDb


class Crate(SeratoBinDb):
    DEFAULT_DATA = [
        ("vrsn", "1.0/Serato ScratchLive Crate"),
        ("osrt", [("tvcn", "key"), ("brev", "\x00")]),
        ("ovct", [("tvcn", "song"), ("tvcw", "0")]),
        ("ovct", [("tvcn", "playCount"), ("tvcw", "0")]),
        ("ovct", [("tvcn", "artist"), ("tvcw", "0")]),
        ("ovct", [("tvcn", "bpm"), ("tvcw", "0")]),
        ("ovct", [("tvcn", "key"), ("tvcw", "0")]),
        ("ovct", [("tvcn", "album"), ("tvcw", "0")]),
        ("ovct", [("tvcn", "length"), ("tvcw", "0")]),
        ("ovct", [("tvcn", "comment"), ("tvcw", "0")]),
        ("ovct", [("tvcn", "added"), ("tvcw", "0")]),
    ]

    StructType = list[tuple[str, "ValueType"]]
    ValueType = StructType | str | bytes

    ValueOrNoneType = Union[ValueType, None]

    def __init__(self, fname):
        self.data: Crate.StructType = []

        self.path: str = os.path.dirname(os.path.abspath(fname))
        self.filename: str = os.path.basename(fname)

        # Omit the _Serato_ and Subcrates folders at the end
        self.track_path: str = os.path.join(*Crate._split_path(self.path)[:-2])

        if os.path.exists(fname):
            self.load_from_file(fname)
        else:
            logging.error(f"file does not exist: {fname}. Using default data.")
            self.data = Crate.DEFAULT_DATA

    def __str__(self):
        tracks = self.tracks()
        return f"Crate containing {len(tracks)} tracks: \n{'\n'.join(tracks)}"

    @staticmethod
    def _split_path(path: str):
        allparts = []
        while True:
            parts = os.path.split(path)
            if parts[0] == path:  # sentinel for absolute paths
                allparts.insert(0, parts[0])
                break
            elif parts[1] == path:  # sentinel for relative paths
                allparts.insert(0, parts[1])
                break
            else:
                path = parts[0]
                allparts.insert(0, parts[1])
        return allparts

    @staticmethod
    def _get_track_name(value: ValueType) -> str:
        if not isinstance(value, list):
            raise TypeError(f"{Crate.TRACK_FIELD} should be list")
        track_name = value[0][1]
        if not isinstance(track_name, str):
            raise TypeError(f"value should be str")
        return track_name

    def tracks(self) -> list[str]:
        track_names: list[str] = []
        for dat in self.data:
            if dat[0] == Crate.TRACK_FIELD:
                track_name = Crate._get_track_name(dat[1])
                track_names.append(track_name)
        return track_names

    def remove_track(self, track_name: str):
        # track_name needs to include the containing folder name.
        found = False
        for i, dat in enumerate(self.data):
            if dat[0] == Crate.TRACK_FIELD:
                crate_track_name = Crate._get_track_name(dat[1])
                if crate_track_name == track_name:
                    self.data.pop(i)
                    found = True

        if not found:
            raise ValueError(f"Track not found in crate: {track_name}")

    def add_track(self, track_name: str):
        # track name must include the containing folder name
        track_name = os.path.relpath(
            os.path.join(self.track_path, track_name), self.track_path
        )

        if track_name in self.tracks():
            return

        self.data.append((Crate.TRACK_FIELD, [("ptrk", track_name)]))

    def include_tracks_from_folder(self, folder_path: str, replace: bool = False):
        folder_tracks = os.listdir(folder_path)
        folder_tracks = [
            os.path.join(os.path.split(folder_path)[1], track)
            for track in folder_tracks
        ]

        if replace:
            for track in self.tracks():
                if track not in folder_tracks:
                    self.remove_track(track)

        for mfile in folder_tracks:
            self.add_track(mfile)

    @overload
    @staticmethod
    def _parse(data: bytes, field: None = None) -> StructType: ...
    @overload
    @staticmethod
    def _parse(data: bytes, field: str) -> ValueType: ...
    @staticmethod
    def _parse(data: bytes, field: str | None = None) -> ValueType | StructType:
        if not isinstance(data, bytes):
            raise Crate.DataTypeError(data, bytes, field)

        type_id: str | None = Crate._get_type(field) if field else "o"
        if type_id == "o":  # struct
            ret_data: Crate.StructType = []
            i = 0
            while i < len(data):
                field = data[i : i + 4].decode("ascii")
                length = struct.unpack(">I", data[i + 4 : i + 8])[0]
                value = data[i + 8 : i + 8 + length]
                value = Crate._parse(value, field=field)
                ret_data.append((field, value))
                i += 8 + length
            return ret_data
        elif type_id in ("p", "t"):  # text
            return data.decode("utf-16-be")
        elif type_id == "b":  # single byte
            return data
        elif type_id == "u":  # unsigned int
            ret_val: bytes = struct.unpack(">I", data)[0]
            return ret_val
        else:
            raise ValueError(f"unexpected type for field: {field}")

    @staticmethod
    def _dump(data: ValueType, field: str | None = None) -> bytes:
        type_id: str | None = Crate._get_type(field) if field else "o"
        if type_id == "o":  # struct
            if not isinstance(data, list):
                raise Crate.DataTypeError(data, list, field)
            ret_data = bytes()
            for dat in data:
                field = dat[0]
                value = Crate._dump(dat[1], field=field)
                length = struct.pack(">I", len(value))
                ret_data = ret_data + field.encode("utf-8") + length + value
            return ret_data
        elif type_id in ("p", "t"):  # text
            if not isinstance(data, str):
                raise Crate.DataTypeError(data, str, field)
            return data.encode("utf-16-be")
        elif type_id == "b":  # single byte, is a boolean
            # if isinstance(data, str) return data.encode("utf-8") # from imported code, but have never seen this happen.
            if not isinstance(data, bytes):
                raise Crate.DataTypeError(data, bytes, field)
            return data
        elif type_id == "u":  #  unsigned
            if not isinstance(data, bytes):
                raise Crate.DataTypeError(data, bytes, field)
            return struct.pack(">I", data)
        else:
            raise ValueError(f"unexpected type for field: {field}")

    def load_from_file(self, fname: str):
        with open(fname, "rb") as mfile:
            self.data = Crate._parse(mfile.read())

    def save_to_file(self, fname: str | None = None):
        if fname is None:
            fname = os.path.join(self.path, self.filename)

        enc_data = Crate._dump(self.data)
        with open(fname, "wb") as mfile:
            mfile.write(enc_data)

    class EntryDict(TypedDict):
        field: str
        field_name: str
        value: Union["Crate.ValueType", list["Crate.EntryDict"]]

    def to_dicts(self) -> Generator[EntryDict, None, None]:
        for field, value in self.data:
            if isinstance(value, list):
                try:
                    new_val: list[Crate.EntryDict] = [
                        {
                            "field": f,
                            "field_name": Crate.get_field_name(f),
                            "value": v,
                        }
                        for f, v in value
                    ]
                except:
                    print(f"error on {value}")
                    raise
                value = new_val
            else:
                value = repr(value)

            yield {
                "field": field,
                "field_name": Crate.get_field_name(field),
                "value": value,
            }

    def print_data(self):
        for entry in self.to_dicts():
            if isinstance(entry["value"], list):
                line = f"{entry['field']} ({entry['field_name']}): "
                field_lines = []
                for e in entry["value"]:
                    if isinstance(e, tuple):
                        raise TypeError("unexpected type")
                    field_lines.append(
                        f"[ {e['field']} ({e['field_name']}): {e['value']} ]"
                    )
                line += ", ".join(field_lines)
                print(line)
            else:
                print(f"{entry['field']} ({entry['field_name']}): {entry['value']}")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    parser.add_argument("-f", "--filenames_only", action="store_true")
    parser.add_argument("-d", "--data", action="store_true")
    parser.add_argument(
        "-o", "--output", "--output_file", dest="output_file", default=None
    )
    args = parser.parse_args()

    crate = Crate(args.file)
    tracks = crate.tracks()
    if args.filenames_only:
        track_names = [
            os.path.splitext(os.path.basename(track))[0] for track in crate.tracks()
        ]
        print("\n".join(track_names))
    elif args.data:
        crate.print_data()
    else:
        print(crate)

    if args.output_file:
        crate.save_to_file(args.output_file)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import io
import logging
import os
import struct
import sys
from typing import Callable, Generator, Iterable, NotRequired, TypedDict, Union

if __package__ is None:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from serato_tools.utils.database import SeratoBinDb


class DatabaseV2(SeratoBinDb):
    DEFAULT_DATABASE_FILE = os.path.join(os.path.expanduser("~"), "Music\\_Serato_\\database V2")  # type: ignore

    ValueType = bytes | str | int | tuple  # TODO: improve the tuple
    ParsedType = tuple[str, int, ValueType]

    ValueOrNoneType = Union[ValueType, None]

    def __init__(self, filepath: str = DEFAULT_DATABASE_FILE):
        if not os.path.exists(filepath):
            raise FileNotFoundError(f"file does not exist: {filepath}")

        self.filepath: str = os.path.abspath(filepath)

        self.data_bin: bytes
        self.data: Iterable[DatabaseV2.ParsedType]

        self.parse()

    def __str__(self):
        return str(list(self.to_dicts()))

    def parse(self, fp: io.BytesIO | io.BufferedReader | str | None = None):
        if fp is None:
            fp = self.filepath
        if isinstance(fp, str):
            fp = open(fp, "rb")

        self.data_bin = fp.read()
        fp.seek(0)
        self.data = self._parse_data_item(fp)

    @staticmethod
    def _parse_data_item(fp: io.BytesIO | io.BufferedReader) -> Generator[ParsedType]:
        for header in iter(lambda: fp.read(8), b""):
            assert len(header) == 8
            field_ascii: bytes
            length: int
            field_ascii, length = struct.unpack(">4sI", header)
            field: str = field_ascii.decode("ascii")
            type_id: str = DatabaseV2._get_type(field)

            data = fp.read(length)
            assert len(data) == length

            value: bytes | str | tuple
            if type_id in ("o", "r"):  #  struct
                value = tuple(DatabaseV2._parse_data_item(io.BytesIO(data)))
            elif type_id in ("p", "t"):  # text
                # value = (data[1:] + b"\00").decode("utf-16") # from imported code
                value = data.decode("utf-16-be")
            elif type_id == "b":  # single byte, is a boolean
                value = struct.unpack("?", data)[0]
            elif type_id == "s":  # signed int
                value = struct.unpack(">H", data)[0]
            elif type_id == "u":  # unsigned int
                value = struct.unpack(">I", data)[0]
            else:
                raise ValueError(f"unexpected type for field: {field}")

            yield field, length, value

    class ModifyRule(TypedDict):
        field: str
        func: Callable[
            [str, "DatabaseV2.ValueOrNoneType"], "DatabaseV2.ValueOrNoneType"
        ]
        """ (filename: str, prev_value: ValueType | None) -> new_value: ValueType | None """
        files: NotRequired[list[str]]

    def modify(self, rules: list[ModifyRule], print_changes: bool = True):
        output = io.BytesIO()
        DatabaseV2._modify_data_item(output, list(self.data), rules, print_changes)
        output.seek(0)
        self.data_bin = output.getvalue()
        output.seek(0)
        self.data = self._parse_data_item(output)

    @staticmethod
    def _modify_data_item(
        fp: io.BytesIO | io.BufferedWriter,
        item: Iterable[ParsedType],
        rules: list[ModifyRule],
        print_changes: bool = True,
    ):
        DatabaseV2._check_rule_fields(rules)

        for rule in rules:
            rule["field_found"] = False  # type: ignore
            if "files" in rule:
                rule["files"] = [
                    DatabaseV2.remove_drive_from_filepath(file)[0].upper()
                    for file in rule["files"]
                ]

        def _dump(field: str, value: "DatabaseV2.ValueType"):
            nonlocal rules, print_changes
            field_bytes = field.encode("ascii")
            assert len(field_bytes) == 4

            type_id: str = DatabaseV2._get_type(field)

            if type_id in ("o", "r"):  #  struct
                if not isinstance(value, tuple):
                    raise DatabaseV2.DataTypeError(value, tuple, field)
                nested_buffer = io.BytesIO()
                DatabaseV2._modify_data_item(nested_buffer, value, rules, print_changes)
                data = nested_buffer.getvalue()
            elif type_id in ("p", "t"):  # text
                if not isinstance(value, str):
                    raise DatabaseV2.DataTypeError(value, str, field)
                # if this ever fails, we did used to do this a different way, see old commits.
                data = value.encode("utf-16-be")
            elif type_id == "b":  # single byte, is a boolean
                if not isinstance(value, bool):
                    raise DatabaseV2.DataTypeError(value, bool, field)
                data = struct.pack("?", value)
            elif type_id == "s":  # signed int
                if not isinstance(value, int):
                    raise DatabaseV2.DataTypeError(value, int, field)
                data = struct.pack(">H", value)
            elif type_id == "u":  # unsigned int
                if not isinstance(value, int):
                    raise DatabaseV2.DataTypeError(value, int, field)
                data = struct.pack(">I", value)
            else:
                raise ValueError(f"unexpected type for field: {field}")

            length = len(data)
            header = struct.pack(">4sI", field_bytes, length)
            fp.write(header)
            fp.write(data)

        def _maybe_perform_rule(
            rule: DatabaseV2.ModifyRule,
            field: str,
            prev_val: "DatabaseV2.ValueOrNoneType",
        ):
            nonlocal track_filename
            if track_filename == "" or (
                "files" in rule and track_filename.upper() not in rule["files"]
            ):
                return None

            maybe_new_value = rule["func"](track_filename, prev_val)
            if maybe_new_value is None or maybe_new_value == prev_val:
                return None

            if rule["field"] == "pfil":
                if not isinstance(maybe_new_value, str):
                    raise DatabaseV2.DataTypeError(maybe_new_value, str, rule["field"])
                if not os.path.exists(maybe_new_value):
                    raise FileNotFoundError(
                        f"set track location to {maybe_new_value}, but doesn't exist"
                    )
                maybe_new_value = DatabaseV2.remove_drive_from_filepath(
                    maybe_new_value
                )[0]

            if print_changes:
                field_name = DatabaseV2.get_field_name(field)
                print(
                    f"Set {field}({field_name})={str(maybe_new_value)} in library for {track_filename}"
                )
            return maybe_new_value

        track_filename: str = ""
        for field, length, value in item:
            if field == "pfil":
                assert isinstance(value, str)
                track_filename = os.path.normpath(value)

            rule = next((r for r in rules if field == r["field"]), None)
            if rule:
                rule["field_found"] = True  # type: ignore
                maybe_new_value = _maybe_perform_rule(rule, field, value)
                if maybe_new_value is not None:
                    value = maybe_new_value

            _dump(field, value)

        for rule in rules:
            if not rule["field_found"]:  # type: ignore
                field = rule["field"]
                maybe_new_value = _maybe_perform_rule(rule, field, None)
                if maybe_new_value is not None:
                    _dump(field, maybe_new_value)

    def modify_file(
        self,
        rules: list[ModifyRule],
        out_file: str | None = None,
        print_changes: bool = True,
    ):
        self.modify(rules, print_changes)
        self.write_file(out_file)

    def write_file(self, out_file: str | None = None):
        if out_file is None:
            out_file = self.filepath
        with open(out_file, "wb") as write_file:
            write_file.write(self.data_bin)

    def rename_track_file(self, src: str, dest: str, print_changes: bool = True):
        """
        This renames the file path, and also changes the path in the database to point to the new filename, so that
        the renamed file is not missing in the library.
        """
        try:
            os.rename(src=src, dst=dest)
            if print_changes:
                print(f"renamed {src} to {dest}")
        except FileExistsError:
            # can't just do os.path.exists, doesn't pick up case changes for certain filesystems
            logging.error("File already exists with change", src)
            return
        self.modify_file(
            [{"field": "pfil", "files": [src], "func": lambda *args: dest}],
            print_changes=print_changes,
        )

    class EntryDict(TypedDict):
        field: str
        field_name: str
        value: str | int | bool | list["DatabaseV2.EntryDict"]
        size_bytes: int

    def to_dicts(self) -> Generator[EntryDict, None, None]:
        for field, length, value in self.data:
            if isinstance(value, tuple):
                try:
                    new_val: list[DatabaseV2.EntryDict] = [
                        {
                            "field": f,
                            "field_name": DatabaseV2.get_field_name(f),
                            "size_bytes": l,
                            "value": v,
                        }
                        for f, l, v in value
                    ]
                except:
                    print(f"error on {value}")
                    raise
                value = new_val
            else:
                value = repr(value)

            yield {
                "field": field,
                "field_name": DatabaseV2.get_field_name(field),
                "size_bytes": length,
                "value": value,
            }

    def print_data(self):
        for entry in self.to_dicts():
            if isinstance(entry["value"], list):
                print(
                    f"{entry['field']} ({entry['field_name']}, {entry['size_bytes']} B)"
                )
                for e in entry["value"]:
                    print(
                        f"    {e['field']} ({e['field_name']}, {e['size_bytes']} B): {e['value']}"
                    )
            else:
                print(
                    f"{entry['field']} ({entry['field_name']}, {entry['size_bytes']} B): {entry['value']}"
                )

    def find_missing(self, drive_letter: str | None = None):
        raise NotImplementedError("TODO: debug. This currently ruins the database.")
        if drive_letter is None:
            drive_letter = os.path.splitdrive(self.filepath)[0]

        track_filepath: str = ""
        missing_checked = False

        def take_input_and_change_db(current_filepath: str):
            while True:
                new_filepath = input("Enter new filepath, or s to skip:")
                new_filepath = new_filepath.strip().strip('"').strip()
                if new_filepath.lower() == "s":
                    return
                if os.path.exists(new_filepath):
                    break
                else:
                    print(f"entered file does not exist: {new_filepath}")

            self.modify(
                [
                    {
                        "field": "pfil",
                        "files": [current_filepath],
                        "func": lambda *args: new_filepath,
                    },
                    {
                        "field": "bmis",
                        "files": [current_filepath],
                        "func": lambda *args: False,
                    },
                ],
                print_changes=True,
            )
            print("\n")

        def field_actions(entry: DatabaseV2.EntryDict):
            nonlocal track_filepath
            nonlocal missing_checked
            if entry["field"] == "pfil":
                if not isinstance(entry["value"], str):
                    raise DatabaseV2.DataTypeError(entry["value"], str, entry["field"])

                value = os.path.join(drive_letter + "//", entry["value"])
                track_filepath = value
                missing_checked = False

                if not os.path.exists(value):
                    print(f"file does not exist: {value}")
                    take_input_and_change_db(current_filepath=entry["value"])
                    missing_checked = True

            elif (not missing_checked) and (entry["field"] == "bmis"):
                missing_checked = True
                if entry["value"]:
                    print(f"serato says is missing: {track_filepath}")
                    take_input_and_change_db(current_filepath=track_filepath)

        for entry in list(self.to_dicts()):
            if isinstance(entry["value"], list):
                for e in entry["value"]:
                    field_actions(e)
            else:
                field_actions(entry)

        self.write_file()


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("file", nargs="?", default=DatabaseV2.DEFAULT_DATABASE_FILE)
    parser.add_argument("--find_missing", action="store_true")
    args = parser.parse_args()

    db = DatabaseV2(args.file)
    if args.find_missing:
        db.find_missing()
    else:
        db.print_data()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import ast
import base64
import configparser
import io
import logging
import os
import struct
import sys
from typing import Callable, NotRequired, Tuple, TypedDict

if __package__ is None:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from mutagen.mp3 import MP3

from serato_tools.utils.track_tags import (
    VERSION_FORMAT,
    check_version,
    del_geob,
    get_geob,
    pack_version,
    tag_geob,
)

from serato_tools import track_cues_v1

GEOB_KEY = "Serato Markers2"

VERSION_BYTES = (0x01, 0x01)

CUE_COLORS = {
    k: bytes.fromhex(v)
    for k, v in {
        "red": "CC0000",  # Hot Cue 1
        "orange": "CC4400",
        "yelloworange": "CC8800",  # Hot Cue 2
        "yellow": "CCCC00",  # Hot Cue 4
        "limegreen1": "88CC00",
        "darkgreen": "44CC00",
        "limegreen2": "00CC00",  # Hot Cue 5
        "limegreen3": "00CC44",
        "seafoam": "00CC88",
        "cyan": "00CCCC",  # Hot Cue 7
        "lightblue": "0088CC",
        "blue1": "0044CC",
        "blue2": "0000CC",  # Hot Cue 3
        "purple1": "4400CC",
        "purple2": "8800CC",  # Hot Cue 8
        "pink": "CC00CC",  # Hot Cue 6
        "magenta": "CC0088",
        "pinkred": "CC0044",
    }.items()
}

TRACK_COLORS = {
    k: bytes.fromhex(v)
    for k, v in {
        "pink": "FF99FF",
        "darkpink": "FF99DD",
        "pinkred": "FF99BB",
        "red": "FF9999",
        "orange": "FFBB99",
        "yelloworange": "FFDD99",
        "yellow": "FFFF99",
        "limegreen1": "DDFF99",
        "limegreen2": "BBFF99",
        "limegreen3": "99FF99",
        "limegreen4": "99FFBB",
        "seafoam": "99FFDD",
        "cyan": "99FFFF",
        "lightblue": "99DDFF",
        "blue1": "99BBFF",
        "blue2": "9999FF",
        "purple": "BB99FF",
        "magenta": "DD99FF",
        "white": "FFFFFF",
        "grey": "BBBBBB",
        "black": "999999",
    }.items()
}


def get_cue_color_key(value: bytes) -> str | None:
    for key, v in CUE_COLORS.items():
        if v == value:
            return key
    return None


def get_track_color_key(value: bytes) -> str | None:
    for key, v in TRACK_COLORS.items():
        if v == value:
            return key
    return None


def readbytes(fp: io.BytesIO):
    for x in iter(lambda: fp.read(1), b""):
        if x == b"\00":
            break
        yield x


class Entry(object):
    NAME: str | None
    FIELDS: Tuple[str, ...]
    data: bytes

    def __init__(self, *args):
        assert len(args) == len(self.FIELDS)
        for field, value in zip(self.FIELDS, args):
            setattr(self, field, value)

    def __repr__(self):
        return "{name}({data})".format(
            name=self.__class__.__name__,
            data=", ".join(
                "{}={!r}".format(name, getattr(self, name)) for name in self.FIELDS
            ),
        )

    @classmethod
    def load(cls, data: bytes):
        return cls(data)

    def dump(self) -> bytes:
        return self.data


class UnknownEntry(Entry):
    NAME = None
    FIELDS = ("data",)

    @classmethod
    def load(cls, data: bytes):
        return cls(data)

    def dump(self):
        return self.data


class BpmLockEntry(Entry):
    NAME = "BPMLOCK"
    FIELDS = ("enabled",)
    FORMAT = "?"

    @classmethod
    def load(cls, data: bytes):
        return cls(*struct.unpack(cls.FORMAT, data))

    def dump(self):
        return struct.pack(self.FORMAT, *(getattr(self, f) for f in self.FIELDS))


class ColorEntry(Entry):
    NAME = "COLOR"
    FORMAT = "c3s"
    FIELDS = (
        "field1",
        "color",
    )

    @classmethod
    def load(cls, data: bytes):
        return cls(*struct.unpack(cls.FORMAT, data))

    def dump(self):
        return struct.pack(self.FORMAT, *(getattr(self, f) for f in self.FIELDS))


class CueEntry(Entry):
    NAME = "CUE"
    FORMAT = ">cBIc3s2s"
    FIELDS = (
        "field1",
        "index",
        "position",
        "field4",
        "color",
        "field6",
        "name",
    )
    name: str

    @classmethod
    def load(cls, data: bytes):
        info_size = struct.calcsize(cls.FORMAT)
        info = struct.unpack(cls.FORMAT, data[:info_size])
        name, nullbyte, other = data[info_size:].partition(b"\x00")
        assert nullbyte == b"\x00"
        assert other == b""
        return cls(*info, name.decode("utf-8"))

    def dump(self):
        struct_fields = self.FIELDS[:-1]
        return b"".join(
            (
                struct.pack(self.FORMAT, *(getattr(self, f) for f in struct_fields)),
                self.name.encode("utf-8"),
                b"\x00",
            )
        )


class LoopEntry(Entry):
    NAME = "LOOP"
    FORMAT = ">cBII4s4sB?"
    FIELDS = (
        "field1",
        "index",
        "startposition",
        "endposition",
        "field5",
        "field6",
        "color",
        "locked",
        "name",
    )
    name: str

    @classmethod
    def load(cls, data: bytes):
        info_size = struct.calcsize(cls.FORMAT)
        info = struct.unpack(cls.FORMAT, data[:info_size])
        name, nullbyte, other = data[info_size:].partition(b"\x00")
        assert nullbyte == b"\x00"
        assert other == b""
        return cls(*info, name.decode("utf-8"))

    def dump(self):
        struct_fields = self.FIELDS[:-1]
        return b"".join(
            (
                struct.pack(self.FORMAT, *(getattr(self, f) for f in struct_fields)),
                self.name.encode("utf-8"),
                b"\x00",
            )
        )


class FlipEntry(Entry):
    NAME = "FLIP"
    FORMAT1 = "cB?"
    FORMAT2 = ">BI"
    FORMAT3 = ">BI16s"
    FIELDS = ("field1", "index", "enabled", "name", "loop", "num_actions", "actions")

    @classmethod
    def load(cls, data):
        info1_size = struct.calcsize(cls.FORMAT1)
        info1 = struct.unpack(cls.FORMAT1, data[:info1_size])
        name, nullbyte, other = data[info1_size:].partition(b"\x00")
        assert nullbyte == b"\x00"

        info2_size = struct.calcsize(cls.FORMAT2)
        loop, num_actions = struct.unpack(cls.FORMAT2, other[:info2_size])
        action_data = other[info2_size:]
        actions = []
        for i in range(num_actions):
            type_id, size = struct.unpack(cls.FORMAT2, action_data[:info2_size])
            action_data = action_data[info2_size:]
            if type_id == 0:
                payload = struct.unpack(">dd", action_data[:size])
                actions.append(("JUMP", *payload))
            elif type_id == 1:
                payload = struct.unpack(">ddd", action_data[:size])
                actions.append(("CENSOR", *payload))
            action_data = action_data[size:]
        assert action_data == b""

        return cls(*info1, name.decode("utf-8"), loop, num_actions, actions)

    def dump(self):
        raise NotImplementedError("FLIP entry dumps are not implemented!")


def get_entry_type(entry_name: str):
    for entry_cls in (BpmLockEntry, ColorEntry, CueEntry, LoopEntry, FlipEntry):
        if entry_cls.NAME == entry_name:
            return entry_cls
    return UnknownEntry


def parse(data: bytes):
    versionlen = struct.calcsize(VERSION_FORMAT)
    check_version(data[:versionlen], VERSION_BYTES)

    try:
        b64data = data[versionlen : data.index(b"\x00", versionlen)]
    except:
        b64data = data[versionlen:]
    b64data = b64data.replace(b"\n", b"")
    padding = b"A==" if len(b64data) % 4 == 1 else (b"=" * (-len(b64data) % 4))
    payload = base64.b64decode(b64data + padding)
    fp = io.BytesIO(payload)
    check_version(fp.read(2), VERSION_BYTES)
    while True:
        entry_name = b"".join(readbytes(fp)).decode("utf-8")
        if not entry_name:
            break
        entry_len = struct.unpack(">I", fp.read(4))[0]
        assert entry_len > 0

        entry_type = get_entry_type(entry_name)
        yield entry_type.load(fp.read(entry_len))


def dump(entries: list[Entry]):
    version = pack_version(VERSION_BYTES)

    contents = [version]
    for entry in entries:
        if entry.NAME is None:
            contents.append(entry.dump())
        else:
            data = entry.dump()
            contents.append(
                b"".join(
                    (
                        entry.NAME.encode("utf-8"),
                        b"\x00",
                        struct.pack(">I", (len(data))),
                        data,
                    )
                )
            )

    payload = b"".join(contents)
    payload_base64 = bytearray(base64.b64encode(payload).replace(b"=", b"A"))

    i = 72
    while i < len(payload_base64):
        payload_base64.insert(i, 0x0A)
        i += 73

    data = version
    data += payload_base64
    return data.ljust(470, b"\x00")


def parse_entries_file(contents: str, assert_len_1: bool):
    cp = configparser.ConfigParser()
    cp.read_string(contents)
    sections = tuple(sorted(cp.sections()))
    if assert_len_1:
        assert len(sections) == 1

    results: list[Entry] = []
    for section in sections:
        l, s, r = section.partition(": ")
        entry_type = get_entry_type(r if s else l)

        e = entry_type(
            *(
                ast.literal_eval(
                    cp.get(section, field),
                )
                for field in entry_type.FIELDS
            )
        )
        results.append(entry_type.load(e.dump()))
    return results


ValueType = bytes | str


class EntryModifyRule(TypedDict):
    field: str
    func: Callable[[ValueType], ValueType | None]
    """ (filename: str, prev_value: ValueType) -> new_value: ValueType | None """


def modify_entry(
    entry: Entry,
    rules: list[EntryModifyRule],
    print_changes: bool = True,
):
    """
    Returns:
        entry: entry if was modified. If was not changed, returns None.
    """

    all_field_names = [rule["field"] for rule in rules]
    assert len(rules) == len(
        list(set(all_field_names))
    ), f"must only have 1 function per field. fields passed: {str(sorted(all_field_names))}"
    # TODO: ensure field is valid else throw error!

    change_made = False

    output = f"[{entry.NAME}]\n"
    for field in entry.FIELDS:
        value: ValueType = getattr(entry, field)

        rule = next((r for r in rules if field == r["field"]), None)
        if rule:
            maybe_new_val = rule["func"](value)
            if maybe_new_val is not None and maybe_new_val != value:
                value = maybe_new_val
                change_made = True
                if print_changes:
                    if isinstance(entry, ColorEntry):
                        color_name = (
                            get_track_color_key(value)
                            if isinstance(value, bytes)
                            else None
                        )
                        print(
                            f"Set Track Color to {color_name if color_name else f'Unknown Color ({str(value)})'}"
                        )
                    elif field == "color":
                        color_name = (
                            get_cue_color_key(value)
                            if isinstance(value, bytes)
                            else None
                        )
                        print(
                            f"Set Cue Color to {color_name if color_name else f'Unknown Color ({str(value)})'}"
                        )
                    else:
                        print(f'Set cue entry field "{field}" to {str(value)}')

        output += f"{field}: {value!r}\n"
    output += "\n"

    if not change_made:
        return None

    entry = parse_entries_file(output, assert_len_1=True)[0]
    return entry


class EntryModifyRules(TypedDict):
    cues: NotRequired[list[EntryModifyRule]]
    color: NotRequired[list[EntryModifyRule]]


def modify_file_entries(
    file: str | MP3,
    rules: EntryModifyRules,
    print_changes: bool = True,
    delete_tags_v1: bool = True,
):
    """
    Args:
        delete_tags_v1: Must delete delete_tags_v1 in order for many tags_v2 changes appear in Serato (since we never change tags_v1 along with it (TODO)). Not sure what tags_v1 is even for, probably older versions of Serato. Have found no issues with deleting this, but use with caution if running an older version of Serato.
    """
    if isinstance(file, str):
        try:
            tags = MP3(file)
        except:
            logging.error("Mutagen error for file %s" % file)
            raise
    else:
        tags = file

    assert tags.filename, "must have filename"

    try:
        data = get_geob(tags, GEOB_KEY)
    except KeyError:
        logging.debug(
            f'File is missing "{GEOB_KEY}" tag, no cue points set yet', tags.filename
        )
        return

    if delete_tags_v1:
        del_geob(tags, track_cues_v1.GEOB_KEY)

    entries = list(parse(data))

    new_entries = []
    change_made = False
    for entry in entries:
        maybe_new_entry = None
        if "cues" in rules and isinstance(entry, CueEntry):
            maybe_new_entry = modify_entry(entry, rules["cues"], print_changes)
        elif "color" in rules and isinstance(entry, ColorEntry):
            maybe_new_entry = modify_entry(entry, rules["color"], print_changes)
        if maybe_new_entry is not None:
            entry = maybe_new_entry
            change_made = True
        new_entries.append(entry)

    if not change_made:
        return

    new_data = dump(new_entries)
    tag_geob(tags, GEOB_KEY, new_data)
    tags.save()


# TODO: allow color or key of colors dict
def set_track_color(
    file: str | MP3,
    color: bytes,
    print_changes: bool = True,
    delete_tags_v1: bool = True,
):
    """
    Args:
        delete_tags_v1: Must delete delete_tags_v1 in order for track color change to appear in Serato (since we never change tags_v1 along with it (TODO)). Not sure what tags_v1 is even for, probably older versions of Serato. Have found no issues with deleting this, but use with caution if running an older version of Serato.
    """
    modify_file_entries(
        file,
        {"color": [{"field": "color", "func": lambda v: color}]},
        print_changes,
        delete_tags_v1,
    )


def is_beatgrid_locked(entries: list[Entry]):
    any(
        (isinstance(entry, BpmLockEntry) and getattr(entry, "enabled"))
        for entry in entries
    )


if __name__ == "__main__":
    import argparse
    import math
    import subprocess
    import tempfile

    import mutagen._file

    from serato_tools.utils.ui import get_hex_editor, get_text_editor, ui_ask

    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    parser.add_argument(
        "--set_color",
        dest="set_color",
        default=None,
        help="Set track color",
    )
    parser.add_argument("-e", "--edit", action="store_true")
    args = parser.parse_args()

    if args.set_color:
        if args.set_color in TRACK_COLORS:
            set_track_color(args.file, TRACK_COLORS[args.set_color], print_changes=True)
        else:
            print(f"Track color must be one of: {str(list(TRACK_COLORS.keys()))}")
        sys.exit()

    if args.edit:
        text_editor = get_text_editor()
        hex_editor = get_hex_editor()

    tagfile = mutagen._file.File(args.file)
    if tagfile is not None:
        data = get_geob(tagfile, GEOB_KEY)
    else:
        with open(args.file, mode="rb") as fp:
            data = fp.read()

    entries = list(parse(data))
    new_entries: list[Entry] = []
    action = None

    width = math.floor(math.log10(len(entries))) + 1
    for entry_index, entry in enumerate(entries):
        if args.edit:
            if action not in ("q", "_"):
                print("{:{}d}: {!r}".format(entry_index, width, entry))
                action = ui_ask(
                    "Edit this entry",
                    {
                        "y": "edit this entry",
                        "n": "do not edit this entry",
                        "q": (
                            "quit; do not edit this entry or any of the "
                            "remaining ones"
                        ),
                        "a": "edit this entry and all later entries in the file",
                        "b": "edit raw bytes",
                        "r": "remove this entry",
                    },
                    default="n",
                )

            if action in ("y", "a", "b"):
                while True:
                    with tempfile.NamedTemporaryFile() as f:
                        if action == "b":
                            f.write(entry.dump())
                            editor = hex_editor
                        else:
                            if action == "a":
                                entries_to_edit = (
                                    (
                                        "{:{}d}: {}".format(i, width, e.NAME),
                                        e,
                                    )
                                    for i, e in enumerate(
                                        entries[entry_index:], start=entry_index
                                    )
                                )
                            else:
                                entries_to_edit = ((entry.NAME, entry),)

                            for section, e in entries_to_edit:
                                f.write("[{}]\n".format(section).encode())
                                for field in e.FIELDS:
                                    f.write(
                                        "{}: {!r}\n".format(
                                            field,
                                            getattr(e, field),
                                        ).encode()
                                    )
                                f.write(b"\n")
                            editor = text_editor
                        f.flush()
                        status = subprocess.call((editor, f.name))
                        f.seek(0)
                        output = f.read()

                    if status != 0:
                        if (
                            ui_ask(
                                "Command failed, retry",
                                {
                                    "y": "edit again",
                                    "n": "leave unchanged",
                                },
                            )
                            == "n"
                        ):
                            break
                    else:
                        try:
                            if action != "b":
                                results = parse_entries_file(
                                    output.decode(), assert_len_1=action != "a"
                                )
                            else:
                                results = [entry.load(output)]
                        except Exception as e:
                            print(str(e))
                            if (
                                ui_ask(
                                    "Content seems to be invalid, retry",
                                    {
                                        "y": "edit again",
                                        "n": "leave unchanged",
                                    },
                                )
                                == "n"
                            ):
                                break
                        else:
                            for i, e in enumerate(results, start=entry_index):
                                print("{:{}d}: {!r}".format(i, width, e))
                            subaction = ui_ask(
                                "Above content is valid, save changes",
                                {
                                    "y": "save current changes",
                                    "n": "discard changes",
                                    "e": "edit again",
                                },
                                default="y",
                            )
                            if subaction == "y":
                                new_entries.extend(results)
                                if action == "a":
                                    action = "_"
                                break
                            elif subaction == "n":
                                if action == "a":
                                    action = "q"
                                new_entries.append(entry)
                                break
            elif action in ("r", "_"):
                continue
            else:
                new_entries.append(entry)
        else:
            print("{:{}d}: {!r}".format(entry_index, width, entry))

    if args.edit:
        if new_entries == entries:
            print("No changes made.")
        else:
            new_data = dump(new_entries)

            if tagfile is not None:
                tag_geob(tagfile, GEOB_KEY, new_data)
                tagfile.save()
            else:
                with open(args.file, mode="wb") as fp:
                    fp.write(new_data)



from typing import Iterable, TypeVar, Union, cast

T = TypeVar("T")


def to_array(x: Union[T, Iterable[T]]) -> Iterable[T]:
    if isinstance(x, (str, bytes)):
        return cast(list[T], [x])
    if isinstance(x, Iterable):
        return x
    return [x]
